/*
 *	Programme de Mandelbrot sequentiel  :
 *	cree un fichier pgm  visualiser avec eog
 *
 *	Compilation : g++ -g -Wall -std=c++11 image.c mandel-list.cpp -o mandel.exx
 *
 *  Quelques jolies fractales
 *      ./mandel.ex 2000 2000 0.35 0.355 0.353 0.358 200
 *      ./mandel.ex 2000 2000 -0.736 -0.184 -0.735 -0.183 500
 *      ./mandel.ex 2000 2000 -0.736 -0.184 -0.735 -0.183 300
 *      ./mandel.ex 2000 2000 -1.48478 0.00006 -1.48440 0.00044 100
 */


#include <stdlib.h>
#include <stdio.h>
#include <time.h>	/* chronometrage */
#include "image.h"

#include <thread>
#include <mutex>
#include <vector>
#include <unistd.h>

using namespace std;


// *********************************************************************
//
//  Travail  faire, pour chaque ligne on calcule la couleur d'une ligne
//
// *********************************************************************
// description du travail
typedef struct {
  // les coordonne de la zone  afficher
  double xmin;
  double xmax;
  double ymin;
  double ymax;
  // la dimension de l'image
  int dimX;
  int dimY;
  // la profondeur de calcul
  int prof;
} descr;

typedef struct {
  int *tab;
  int nbtaches;
  int debut;
} lt;

lt* new_liste_taches(int nbtaches) {
  int i;
  lt *res = (lt *) malloc(sizeof(lt));

  res->nbtaches = nbtaches;
  res->debut = 0;
  res->tab = (int *) malloc(nbtaches*sizeof(int));
  for (i=0; i<nbtaches; i++) {
    res->tab[i] = i;
  }
  return res;
}

void
delete_liste_taches(lt *l) {
  free(l->tab);
  free(l);
}

// fonction ui retourne une tache (ou -1 si la liste est vide)
int get_tache(lt *l) {
  if (l->debut == l->nbtaches) {
    return -1;
  }
  int res = l->tab[l->debut];
  usleep(10);
  l->debut++;

  return res;
}

void usage(const char* nom) {
  printf("Usage : %s dim0 dim1 Xmin Ymin Xmax Ymax Prof \n", nom);
}

descr lecture_args(int argc, char *argv[]) {
  descr   d;

  /* Valeurs par defaut de la fractale */
  // 2000 2000 -0.736 -0.184 -0.735 -0.183 500
  d.xmin = -0.736;
  d.ymin = -0.184;
  d.xmax = -0.735;
  d.ymax = -0.183;
  d.dimX = 2000;
  d.dimY = 2000;
  d.prof = 500;
  
  /* Recuperation des parametres */
  if( argc > 1) {
    d.dimX = atoi(argv[1]);
    if (d.dimX == 0) {
      usage(argv[0]);
      exit(1);
    }
  }
  if ( argc > 2) {
    d.dimY = atoi(argv[2]);
    if (d.dimY == 0) {
      usage(argv[0]);
      exit(1);
    }
  }
  if (argc > 3) d.xmin = atof(argv[3]);
  if( argc > 4) d.ymin = atof(argv[4]);
  if( argc > 5) d.xmax = atof(argv[5]);
  if( argc > 6) d.ymax = atof(argv[6]);
  if( argc > 7)  {
    d.prof = atoi(argv[7]);
    if (d.prof == 0) {
      usage(argv[0]);
      exit(1);
    }
  }
  if (argc > 8) {
      usage(argv[0]);
      exit(1);
  }

  return d;
}

void calcul_ligne(image_t image, descr d, int numligne) {
  int i;
  int depart = numligne*d.dimY;
  double x = numligne*(d.xmax-d.xmin)/d.dimX+d.xmin;
  double y = d.ymin;
  double pas = (d.ymax-d.ymin)/d.dimY;

  if ((numligne<0)||(numligne>=d.dimX)) {
    fprintf(stderr, "Attention, vous tentez de calculer la ligne %d d'une image qui a %d lignes\n",
	    numligne, d.dimX);
    return;
  }
  
  pixel_t *p = image.data+depart;
  if (*p != 0) {
    fprintf(stderr, "Attention, la ligne %d de l'image image semble avoir dj t calcule (je continue, mais il y a peut-tre une erreur)\n",
	    numligne);    
  }
  for (i=0; i<d.dimY; i++) {
    p[i] = xy2color(x,y, d.prof);
    y += pas;
  }
}



// ********************************************************************************
// 
//   modifier  partir de l
//   Partie principale: pour chaque ligne de l'image appliquer la fonction prcdante
// ********************************************************************************

void *f_thread(int numth, lt *l, image_t &image, descr &d, int &nblignes, mutex &m) {

  fprintf(stdout, "lancement du thread %d\n", numth);

  /* Traitement de l'image ligne par ligne*/
  nblignes = 0;
  while (1) {
    m.lock();    
    int t = get_tache(l);
    m.unlock();

    if (t == -1) {// la liste de taches est termine et dans ce cas la fonction retourne -1
      break;
    }
    //    fprintf(stderr, "thread %d : calcule ligne %d et %x, %x\n", p->numth, t, (p->image), p->image->data);
    calcul_ligne(image, d, t);
    nblignes++;
  }

  return NULL;
}

int main(int argc, char *argv[]) {
  descr   d;
  clock_t debut, fin;		    /* chronometrage */
  int nbth = 10;
  

  
  /* debut du chronometrage */
  debut = clock();

  /* lecture des argument */
  d = lecture_args(argc, argv);
  
  /* affichage parametres pour control */
  printf( "Domaine: %lf, %lf %lf %lf\n", 
	  d.xmin, d.ymin, d.xmax, d.ymax); 
  printf( "Prof: %d\n",  d.prof);
  printf( "Dim image: %d %d\n", d.dimX, d.dimY);
  
  // cration de l'image
  image_t im = image_new(d.dimX, d.dimY);

  // cration de la liste des tches
  lt *l = new_liste_taches(d.dimX);

  int i;
  vector<thread> tab;
  int tabres[nbth];
  mutex m;
  
  // lancement des threads
  for (i=0; i<nbth; i++) {
    tab.push_back(thread(f_thread, i, l, std::ref(im), std::ref(d), std::ref(tabres[i]), std::ref(m)));
  }

  int nblignes = 0;
  for (int i=0; i<nbth; i++) {
    tab[i].join();
    nblignes += tabres[i];
  }
  
  // c'est une vrification pour vous aider a voir les probmes, cela ne servira qu'en cas de multithreading (et si vous parvenez
  // bien  calculer le nombre de lignes fait par chaque threads.
  if (nblignes == d.dimX) {
    fprintf(stdout, "on a bien calcul %d lignes\n", nblignes);
  } else {
    fprintf(stdout, "Attention, %d lignes ont t calcules  la place de %d (ou  le calcul du nombre de ligne est faux)\n", nblignes, d.dimX);
  }
  
  /* Sauvegarde de la grille dans le fichier resultat "result.pgm" */
  image_to_file(im, "result.pgm");
  
  /* fin du chronometrage */
  fin = clock();
  printf("Temps total de calcul : %e seconde \n",
		 (double) (fin - debut)/ CLOCKS_PER_SEC);

  delete_liste_taches(l);
  
  return 0;
}
